import 'dart:convert';
import 'dart:developer';
import 'package:dio/dio.dart';

class NetworkOption extends BaseOptions {
  ///called on start every request
  Function onStartDefault = () {};

  ///called on End every request
  Function onEndDefault = () {};

  ///called on Success every request
  void Function(NetworkResponse) onSuccessDefault = (NetworkResponse response) {};

  ///called on Failed every request
  void Function(NetworkResponse) onFailedDefault = (NetworkResponse response) {};

  CancelToken? globalCancelToken;

  ///String token which is used in header
  String? token;

  ///called on token expire
  bool Function(NetworkResponse) tokenExpireRule = (res) => res.responseCode == 403;
  bool Function(NetworkResponse) extraSuccessRule = (res) => true;
  void Function(NetworkResponse) onTokenExpire = (_) {
    print("Token Expired");
  };

  /// a function to extract error as string
  String Function(dynamic error) errorMsgExtractor = (error) => error.toString();
  String Function(dynamic msg)? successMsgExtractor = (msg) => msg.toString();

  static final NetworkOption _singleton = NetworkOption._internal();

  NetworkOption._internal();

  factory NetworkOption() {
    return _singleton;
  }

  static setAccessToken(String token, {String tokenKey = "Authorization", String tokenPrefix = "Bearer "}) {
    log("Token is set: $token");
    // _singleton.headers.update("$tokenKey", (v) => "$tokenPrefix$token",
    //     ifAbsent: () => _singleton.headers.putIfAbsent("$tokenKey", () => "$tokenPrefix$token"));
    _singleton.token = token;
  }

  static initialize({
    String? baseUrl,
    int timeout = 15000,
    String? token,
    Function? onStartDefault,
    Function? onEndDefault,
    Function? onSuccessDefault,
    Function? onFailedDefault,
    CancelToken? globalCancelToken,
    Map<String, dynamic>? headers,
    bool Function(NetworkResponse)? tokenExpireRule,
    bool Function(NetworkResponse)? extraSuccessRule,
    void Function(NetworkResponse)? onTokenExpire,
    String Function(dynamic error)? errorMsgExtractor,
    String Function(dynamic msg)? successMsgExtractor,
  }) {
    if (onStartDefault != null) {
      _singleton.onStartDefault = onStartDefault;
    }
    if (onEndDefault != null) {
      _singleton.onEndDefault = onEndDefault;
    }
    if (onSuccessDefault != null) {
      _singleton.onSuccessDefault = onSuccessDefault as void Function(NetworkResponse);
    }
    if (onFailedDefault != null) {
      _singleton.onFailedDefault = onFailedDefault as void Function(NetworkResponse);
    }
    if (onTokenExpire != null) {
      _singleton.onTokenExpire = onTokenExpire;
    }
    if (tokenExpireRule != null) {
      _singleton.tokenExpireRule = tokenExpireRule;
    }
    if (extraSuccessRule != null) {
      _singleton.extraSuccessRule = extraSuccessRule;
    }
    if (baseUrl != null) {
      _singleton.baseUrl = baseUrl;
    }
    if (headers != null) {
      _singleton.headers = headers;
    }
    _singleton.connectTimeout = timeout;
    _singleton.sendTimeout = timeout;
    _singleton.receiveTimeout = timeout;

    if (globalCancelToken != null) {
      _singleton.globalCancelToken = globalCancelToken;
    }

    _singleton.token = token;

    ///You can set LogInterceptor to print request/response log automatically
//    _singleton.dio.interceptors.add(LogInterceptor(responseBody: false));
    if (errorMsgExtractor != null) {
      _singleton.errorMsgExtractor = errorMsgExtractor;
    }
    if (successMsgExtractor != null) {
      _singleton.successMsgExtractor = errorMsgExtractor;
    }
  }
}

class NetworkResponse {
  bool responseStatus;
  bool isCancelable;
  bool callDefaults;
  bool handleErrors;
  bool handleSuccess;
  int responseCode;
  String? responseTitle;
  Function? retryFunction;
  dynamic responseBody;
  dynamic responseDetails;
  String? extractedMessage;

  NetworkResponse(
      {required this.responseStatus,
      required this.responseCode,
      this.responseTitle,
      this.responseBody,
      this.isCancelable = true,
      this.callDefaults = true,
      this.handleErrors = true,
      this.handleSuccess = true,
      this.retryFunction,
      this.responseDetails,
      this.extractedMessage});

  @override
  String toString() {
    return "$responseCode $responseTitle : $extractedMessage";
  }
}

class NetworkRequest {
  final String api;
  final String? savePath;
  final dynamic data;
  final Dio dio;
  final Duration timeOut;
  final NetworkOption options;
  final bool callDefaults;
  final bool handleErrors;
  final bool handleSuccess;
  final bool isCancelable;

  final Function? retry;
  final Function(int, int)? onReceiveProgress;

  NetworkRequest._({
    required this.api,
    this.data,
    this.timeOut = const Duration(seconds: 30),
    required this.options,
    this.callDefaults = true,
    this.handleErrors = true,
    this.handleSuccess = true,
    this.isCancelable = true,
    this.retry,
    this.onReceiveProgress,
    this.savePath,
  }) : dio = Dio(options);

  factory NetworkRequest(
      {required api,
      data,
      bool callDefaults = true,
      timeOut = const Duration(seconds: 30),
      bool handleErrors = true,
      bool handleSuccess = true,
      Function? retry,
      void Function(int, int)? onReceiveProgress,
      String? savePath}) {
    NetworkOption option = NetworkOption();
    return NetworkRequest._(
        api: api,
        data: data,
        timeOut: timeOut,
        options: option,
        callDefaults: callDefaults,
        handleErrors: handleErrors ,
        handleSuccess: handleSuccess,
        retry: retry,
        onReceiveProgress: onReceiveProgress,
        savePath: savePath);
  }

  Future<NetworkResponse> post() async {
    if (callDefaults) {
      options.onStartDefault();
    }
    NetworkResponse response;
    try {
      log("Posting ${jsonEncode(data)} in ${jsonEncode(api)}");
      Response res = await dio.post(api, data: data, cancelToken: options.globalCancelToken).catchError((e) {
        throw e;
      }).timeout(timeOut, onTimeout: () {
        throw DioErrorType.connectTimeout;
      });
      response = NetworkResponse(
          responseStatus: res.statusCode! > 199 && res.statusCode! < 300,
          responseCode: res.statusCode!,
          responseTitle: res.statusMessage,
          responseBody: res.data,
          responseDetails: res,
          extractedMessage: options.successMsgExtractor!(res.data));
    } catch (e) {
      if (e is DioError && e.response != null && e.response!.data is Map) {
        response = NetworkResponse(
            responseStatus: e.response!.statusCode! > 199 && e.response!.statusCode! < 300,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: options.errorMsgExtractor(e.response!.data));
      } else if (e is DioError && e.type == DioErrorType.response) {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: "${e.message}");
      } else if (e is DioError) {
        if (e.type == DioErrorType.other) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Connection interrupted ,Can not Connect to Server!");
        } else if (e.type == DioErrorType.connectTimeout ||
            e.type == DioErrorType.receiveTimeout ||
            e.type == DioErrorType.sendTimeout) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection Timeout ,Can not Reach to Server!",
              responseDetails: e,
              extractedMessage: "Connection Timeout ,Can not Reach to Server!");
        } else if (e.type == DioErrorType.cancel) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Request Canceled!",
              responseDetails: e,
              extractedMessage: "Request Canceled!");
        } else {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Connection interrupted ,Can not Connect to Server!");
        }
      } else {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: -500,
            responseTitle: "Failed",
            responseBody: "Connection interrupted ,Can not Connect to Server!",
            responseDetails: e,
            extractedMessage: e.toString());
      }
    }

    response.isCancelable = isCancelable;
    response.callDefaults = callDefaults;
    response.handleSuccess = handleSuccess;
    response.handleErrors = handleErrors;

    response.retryFunction = retry ?? post;
    log(response.toString());

    if (options.tokenExpireRule(response)) {
      options.onTokenExpire(response);
      response.extractedMessage = "Token Expired!";
      response.responseCode = -999;
      response.responseStatus = false;
    }

    if (!options.extraSuccessRule(response)) {
      response.responseStatus = false;
      response.responseCode = -100;
      response.responseTitle = "Success Rules Failed";
    }



    if (callDefaults) {
      if (response.responseStatus) {
        options.onSuccessDefault(response);
      } else {
        options.onFailedDefault(response);
      }
      options.onEndDefault();
    } else {
      if (response.responseStatus) {
        if (handleSuccess) {
          options.onSuccessDefault(response);
        }
      } else {
        if (handleErrors) {
          options.onFailedDefault(response);
        }
      }
    }



    return response;
  }

  Future<NetworkResponse> get() async {
    if (callDefaults) {
      options.onStartDefault();
    }
    NetworkResponse response;
    try {
      String query = api + data;
      log("Getting ${jsonEncode(query)}");
      Response res = await dio.get(query, cancelToken: options.globalCancelToken).catchError((e) {
        throw e;
      }).timeout(timeOut, onTimeout: () {
        throw DioErrorType.connectTimeout;
      });
      response = NetworkResponse(
          responseStatus: res.statusCode! > 199 && res.statusCode! < 300,
          responseCode: res.statusCode!,
          responseTitle: res.statusMessage,
          responseBody: res.data,
          responseDetails: res,
          extractedMessage: options.successMsgExtractor!(res.data));
    } catch (e) {
      if (e is DioError && e.response != null && e.response!.data is Map) {
        response = NetworkResponse(
            responseStatus: e.response!.statusCode! > 199 && e.response!.statusCode! < 300,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: options.errorMsgExtractor(e.response!.data));
      } else if (e is DioError && e.type == DioErrorType.response) {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: "${e.message}");
      } else if (e is DioError) {
        if (e.type == DioErrorType.other) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Connection interrupted ,Can not Connect to Server!");
        } else if (e.type == DioErrorType.connectTimeout ||
            e.type == DioErrorType.receiveTimeout ||
            e.type == DioErrorType.sendTimeout) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection Timeout ,Can not Reach to Server!",
              responseDetails: e,
              extractedMessage: "Connection Timeout ,Can not Reach to Server!");
        } else if (e.type == DioErrorType.cancel) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Request Canceled!",
              responseDetails: e,
              extractedMessage: "Request Canceled!");
        } else {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Connection interrupted ,Can not Connect to Server!");
        }
      } else {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: -500,
            responseTitle: "Failed",
            responseBody: "Connection interrupted ,Can not Connect to Server!",
            responseDetails: e,
            extractedMessage: e.toString());
      }
    }

    response.isCancelable = isCancelable;
    response.callDefaults = callDefaults;
    response.handleSuccess = handleSuccess;
    response.handleErrors = handleErrors;

    response.retryFunction = retry ?? get;
    log(response.toString());

    if (options.tokenExpireRule(response)) {
      options.onTokenExpire(response);
      response.extractedMessage = "Token Expired!";
      response.responseCode = -999;
      response.responseStatus = false;
    }

    if (!options.extraSuccessRule(response)) {
      response.responseStatus = false;
      response.responseCode = -100;
      response.responseTitle = "Success Rules Failed";
    }

    if (callDefaults) {
      if (response.responseStatus) {
        options.onSuccessDefault(response);
      } else {
        options.onFailedDefault(response);
      }
      options.onEndDefault();
    } else {
      if (response.responseStatus) {
        if (handleSuccess) {
          options.onSuccessDefault(response);
        }
      } else {
        if (handleErrors) {
          options.onFailedDefault(response);
        }
      }
    }


    return response;
  }

  Future<NetworkResponse> put() async {
    if (callDefaults) {
      options.onStartDefault();
    }
    NetworkResponse response;
    try {
      log("Putting ${jsonEncode(data)} in ${jsonEncode(api)}");
      Response res = await dio.put(api, data: data, cancelToken: options.globalCancelToken).catchError((e) {
        throw e;
      }).timeout(timeOut, onTimeout: () {
        throw DioErrorType.connectTimeout;
      });
      response = NetworkResponse(
          responseStatus: res.statusCode! > 199 && res.statusCode! < 300,
          responseCode: res.statusCode!,
          responseTitle: res.statusMessage,
          responseBody: res.data,
          responseDetails: res,
          extractedMessage: options.successMsgExtractor!(res.data));
    } catch (e) {
      if (e is DioError && e.response != null && e.response!.data is Map) {
        response = NetworkResponse(
            responseStatus: e.response!.statusCode! > 199 && e.response!.statusCode! < 300,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: options.errorMsgExtractor(e.response!.data));
      } else if (e is DioError && e.type == DioErrorType.response) {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: "${e.message}");
      } else if (e is DioError) {
        if (e.type == DioErrorType.other) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Connection interrupted ,Can not Connect to Server!");
        } else if (e.type == DioErrorType.connectTimeout ||
            e.type == DioErrorType.receiveTimeout ||
            e.type == DioErrorType.sendTimeout) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection Timeout ,Can not Reach to Server!",
              responseDetails: e,
              extractedMessage: "Connection Timeout ,Can not Reach to Server!");
        } else if (e.type == DioErrorType.cancel) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Request Canceled!",
              responseDetails: e,
              extractedMessage: "Request Canceled!");
        } else {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Connection interrupted ,Can not Connect to Server!");
        }
      } else {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: -500,
            responseTitle: "Failed",
            responseBody: "Connection interrupted ,Can not Connect to Server!",
            responseDetails: e,
            extractedMessage: e.toString());
      }
    }


    response.isCancelable = isCancelable;
    response.callDefaults = callDefaults;
    response.handleSuccess = handleSuccess;
    response.handleErrors = handleErrors;

    response.retryFunction = retry ?? put;
    log(response.toString());

    if (options.tokenExpireRule(response)) {
      options.onTokenExpire(response);
      response.extractedMessage = "Token Expired!";
      response.responseCode = -999;
      response.responseStatus = false;
    }

    if (!options.extraSuccessRule(response)) {
      response.responseStatus = false;
      response.responseCode = -100;
      response.responseTitle = "Success Rules Failed";
    }

    if (callDefaults) {
      if (response.responseStatus) {
        options.onSuccessDefault(response);
      } else {
        options.onFailedDefault(response);
      }
      options.onEndDefault();
    } else {
      if (response.responseStatus) {
        if (handleSuccess) {
          options.onSuccessDefault(response);
        }
      } else {
        if (handleErrors) {
          options.onFailedDefault(response);
        }
      }
    }

    return response;
  }

  Future<NetworkResponse> delete() async {
    if (callDefaults) {
      options.onStartDefault();
    }
    NetworkResponse response;
    try {
      log("Deleting ${jsonEncode(data)} from ${jsonEncode(api)}");
      Response res = await dio.delete(api, data: data, cancelToken: options.globalCancelToken).catchError((e) {
        throw e;
      }).timeout(timeOut, onTimeout: () {
        throw DioErrorType.connectTimeout;
      });
      response = NetworkResponse(
          responseStatus: res.statusCode! > 199 && res.statusCode! < 300,
          responseCode: res.statusCode!,
          responseTitle: res.statusMessage,
          responseBody: res.data,
          responseDetails: res,
          extractedMessage: options.successMsgExtractor!(res.data));
    } catch (e) {
      if (e is DioError && e.response != null && e.response!.data is Map) {
        response = NetworkResponse(
            responseStatus: e.response!.statusCode! > 199 && e.response!.statusCode! < 300,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: options.errorMsgExtractor(e.response!.data));
      } else if (e is DioError && e.type == DioErrorType.response) {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: "${e.message}");
      } else if (e is DioError) {
        if (e.type == DioErrorType.other) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Connection interrupted ,Can not Connect to Server!");
        } else if (e.type == DioErrorType.connectTimeout ||
            e.type == DioErrorType.receiveTimeout ||
            e.type == DioErrorType.sendTimeout) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection Timeout ,Can not Reach to Server!",
              responseDetails: e,
              extractedMessage: "Connection Timeout ,Can not Reach to Server!");
        } else if (e.type == DioErrorType.cancel) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Request Canceled!",
              responseDetails: e,
              extractedMessage: "Request Canceled!");
        } else {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Connection interrupted ,Can not Connect to Server!");
        }
      } else {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: -500,
            responseTitle: "Failed",
            responseBody: "Connection interrupted ,Can not Connect to Server!",
            responseDetails: e,
            extractedMessage: e.toString());
      }
    }


    response.isCancelable = isCancelable;
    response.callDefaults = callDefaults;
    response.handleSuccess = handleSuccess;
    response.handleErrors = handleErrors;

    response.retryFunction = retry ?? delete;
    log(response.toString());

    if (options.tokenExpireRule(response)) {
      options.onTokenExpire(response);
      response.extractedMessage = "Token Expired!";
      response.responseCode = -999;
      response.responseStatus = false;
    }

    if (!options.extraSuccessRule(response)) {
      response.responseStatus = false;
      response.responseCode = -100;
      response.responseTitle = "Success Rules Failed";
    }

    if (callDefaults) {
      if (response.responseStatus) {
        options.onSuccessDefault(response);
      } else {
        options.onFailedDefault(response);
      }
      options.onEndDefault();
    } else {
      if (response.responseStatus) {
        if (handleSuccess) {
          options.onSuccessDefault(response);
        }
      } else {
        if (handleErrors) {
          options.onFailedDefault(response);
        }
      }
    }

    return response;
  }

  Future<NetworkResponse> download() async {
    if (callDefaults) {
      options.onStartDefault();
    }
    NetworkResponse response;
    try {
      log("Downloading ${jsonEncode(data)} from ${jsonEncode(api)}");
      Response res = await dio
          .download(api, savePath, onReceiveProgress: onReceiveProgress, data: data, cancelToken: options.globalCancelToken)
          .catchError((e) {
        throw e;
      }).timeout(timeOut, onTimeout: () {
        throw DioErrorType.connectTimeout;
      });
      response = NetworkResponse(
          responseStatus: res.statusCode! > 199 && res.statusCode! < 300,
          responseCode: res.statusCode!,
          responseTitle: res.statusMessage,
          responseBody: res.data,
          responseDetails: res,
          extractedMessage: options.successMsgExtractor!(res.data));
    } catch (e) {
      if (e is DioError && e.response != null && e.response!.data is Map) {
        response = NetworkResponse(
            responseStatus: e.response!.statusCode! > 199 && e.response!.statusCode! < 300,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: options.errorMsgExtractor(e.response!.data));
      } else if (e is DioError && e.type == DioErrorType.response) {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: e.response!.statusCode!,
            responseTitle: e.response!.statusMessage,
            responseBody: e.response!.data,
            responseDetails: e,
            extractedMessage: "${e.message}");
      } else if (e is DioError) {
        if (e.type == DioErrorType.other) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Connection interrupted ,Can not Connect to Server!");
        } else if (e.type == DioErrorType.connectTimeout ||
            e.type == DioErrorType.receiveTimeout ||
            e.type == DioErrorType.sendTimeout) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection Timeout ,Can not Reach to Server!",
              responseDetails: e,
              extractedMessage: "Connection Timeout ,Can not Reach to Server!");
        } else if (e.type == DioErrorType.cancel) {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Request Canceled!",
              responseDetails: e,
              extractedMessage: "Request Canceled!");
        } else {
          response = NetworkResponse(
              responseStatus: false,
              responseCode: -500,
              responseTitle: "Error",
              responseBody: "Connection interrupted ,Can not Connect to Server!",
              responseDetails: e,
              extractedMessage: "Something went wrong");
        }
      } else {
        response = NetworkResponse(
            responseStatus: false,
            responseCode: -500,
            responseTitle: "Failed",
            responseBody: "Connection interrupted ,Can not Connect to Server!",
            responseDetails: e,
            extractedMessage: e.toString());
      }
    }

    response.isCancelable = isCancelable;
    response.callDefaults = callDefaults;
    response.handleSuccess = handleSuccess;
    response.handleErrors = handleErrors;


    response.retryFunction = retry ?? delete;
    log(response.toString());

    if (options.tokenExpireRule(response)) {
      options.onTokenExpire(response);
      response.extractedMessage = "Token Expired!";
      response.responseCode = -999;
      response.responseStatus = false;
    }

    if (!options.extraSuccessRule(response)) {
      response.responseStatus = false;
      response.responseCode = -100;
      response.responseTitle = "Success Rules Failed";
    }

    if (callDefaults) {
      if (response.responseStatus) {
        options.onSuccessDefault(response);
      } else {
        options.onFailedDefault(response);
      }
      options.onEndDefault();
    } else {
      if (response.responseStatus) {
        if (handleSuccess) {
          options.onSuccessDefault(response);
        }
      } else {
        if (handleErrors) {
          options.onFailedDefault(response);
        }
      }
    }

    return response;
  }
}
